# -*- coding: iso-8859-1 -*-
"""
Ein Programm zum Spielen mit dem chi-Quadrat-Test und "Kollokationen".

Vgl. Skript: "Identifikation von Kollokationen"
"""


import operator, math, re, gzip


class Error(Exception):
	pass


try:
	a = sum(range(2))
except NameError:
	def sum(seq):
		return reduce(operator.add, seq, 0)


class DictWithZeroDefault(dict):
	def __getitem__(self, key):
		if self.has_key(key):
			return dict.__getitem__(self, key)
		else:
			return 0


def lzprob(z):
    """ripped from Gary Strangman's stats.py module.  Ignore, used by
    lchisqprob.
    """
    Z_MAX = 6.0    # maximum meaningful z-value
    if z == 0.0:
        x = 0.0
    else:
        y = 0.5 * math.fabs(z)
        if y >= (Z_MAX*0.5):
            x = 1.0
        elif (y < 1.0):
            w = y*y
            x = ((((((((0.000124818987 * w
                        -0.001075204047) * w +0.005198775019) * w
                      -0.019198292004) * w +0.059054035642) * w
                    -0.151968751364) * w +0.319152932694) * w
                  -0.531923007300) * w +0.797884560593) * y * 2.0
        else:
            y = y - 2.0
            x = (((((((((((((-0.000045255659 * y
                             +0.000152529290) * y -0.000019538132) * y
                           -0.000676904986) * y +0.001390604284) * y
                         -0.000794620820) * y -0.002034254874) * y
                       +0.006549791214) * y -0.010557625006) * y
                     +0.011630447319) * y -0.009279453341) * y
                   +0.005353579108) * y -0.002141268741) * y
                 +0.000535310849) * y +0.999936657524
    if z > 0.0:
        prob = ((x+1.0)*0.5)
    else:
        prob = ((1.0-x)*0.5)
    return prob

def lchisqprob(chisq,df):
    """returns the "p-value" for a chi square test with df degrees
    of freedom.  Ignore.  Ripped from Gary Strangman's stat.py
    module.
    """
    BIG = 20.0
    def ex(x):
        BIG = 20.0
        if x < -BIG:
            return 0.0
        else:
            return math.exp(x)

    if chisq <=0 or df < 1:
        return 1.0
    a = 0.5 * chisq
    if df%2 == 0:
        even = 1
    else:
        even = 0
    if df > 1:
        y = ex(-a)
    if even:
        s = y
    else:
        s = 2.0 * lzprob(-math.sqrt(chisq))
    if (df > 2):
        chisq = 0.5 * (df - 1.0)
        if even:
            z = 1.0
        else:
            z = 0.5
        if a > BIG:
            if even:
                e = 0.0
            else:
                e = math.log(math.sqrt(math.pi))
            c = math.log(a)
            while (z <= chisq):
                e = math.log(z) + e
                s = s + ex(c*z-a-e)
                z = z + 1.0
            return s
        else:
            if even:
                e = 1.0
            else:
                e = 1.0 / math.sqrt(math.pi) / math.sqrt(a)
            c = 0.0
            while (z <= chisq):
                e = e * (a/float(z))
                c = c + e
                z = z + 1.0
            return (c*y+s)
    else:
        return s


def _computeExpectations(table):
	"""returns the table of expectations for a chi square contingency
	table.  Used by computeChisquare.
	"""
	marginH = map(lambda a: sum(a), table)
	marginV = [sum([t[col] for t in table]) 
		for col in range(len(table[0]))]
	total = sum(marginH)
	expectations = [[a*b/float(total) for a in marginV] for b in marginH]
	return expectations

def computeChisquare(table, debug=None):
	"""returns the chisquare for the list of lists table together with
	a valid flag that is false if any of the expectations are smaller
	than 2.
	"""
	valid = 1
	if debug:
		print "Beobachtungen:", table
	expectations = _computeExpectations(table)
	if debug:
		print "Erwartungen:", expectations
	if min(map(min, expectations))<2:
		valid = 0
	chiMat = map(lambda row1, row2: map(lambda a,b: (a-b)*(a-b)/b, row1, row2), 
		table, expectations)
	return sum(map(sum, chiMat)), valid

def tokenize(tx):
	"""returns a list of individual words (defined by a regular expression)
	present in the text tx.
	"""
	return re.sub(r"(?L)[^\w\s']", " ", 
			re.sub("(?s)<[^>]*>", "", tx.lower())).split()

def getUnigramFreqs(seq):
	"""returns a dictionary of frequencies of the items in seq.
	"""
	dist = DictWithZeroDefault()
	for item in seq:
		dist[item] = dist.get(item, 0)+1
	return dist

def getBigramFreqs(seq):
	"""returns a dictionary of frequencies of bigrams in seq.
	"""
	return getUnigramFreqs(zip(seq[:-1], seq[1:]))

def _getModels(corpusName):
	"""returns a unigram and a bigram model for the text in corpusName.
	"""
	if corpusName.endswith(".gz"):
		open = gzip.open
	else:
		open = file
	corpus = tokenize(open(corpusName).read())
	return len(corpus), getUnigramFreqs(corpus), getBigramFreqs(corpus)


def runOneTest(head, tail, uniModel, biModel, corpusSize, niveau=0.05):
	chisq, valid = computeChisquare([
			[biModel[head,tail], uniModel[head]-biModel[head,tail]],
			[uniModel[tail]-biModel[head,tail], corpusSize-
				uniModel[head]-uniModel[tail]+biModel[head,tail]]
			], debug=1)
	if not valid:
		print "*** Daten reichen nicht aus, Test ungltig"
	print "Chi Quadrat ist ", chisq
	pVal = lchisqprob(chisq, 1)
	if pVal<niveau:
		rejectString = ""
	else:
		rejectString = "nicht "
	print ("Zum Niveau %f kann Unabhngigkeit %sverworfen werden (pv: %f)"
		)%(niveau, rejectString, pVal)

def runCollocTest(corpusName):
	"""reads a corpus from the file pointed to by corpusName, builds both
	a unigram and a bigram model for it and then prompts the user for
	two words on a line to check for collocation-ness.  It runs a
	chi square-test and complains if not enough data is available.
	"""
	corpusLength, uniModel, biModel = _getModels(corpusName)
	while 1:
		words = raw_input("\nZwei Wrter, leer fr Abbruch? ").split()
		if not len(words)==2:
			break
		try:
			runOneTest(words[0], words[1], uniModel, biModel, corpusLength)
		except ZeroDivisionError:
			print "Eines der Wrter kam nicht vor."
	
def showSortedCollocs(corpusName, niveau=0.05, freqThreshold=500):
	"""prints decisions if bigrams encoutered in the corpus pointed to
	by corpusName are "Collocations" or not.  The most frequent bigrams
	are handled first.  Bigrams with insufficient data are printed in
	parentheses.  You can skip frequent words by setting freqThreshold
	to !=0.  If both words of a bigram are more frequent than
	corpusLength/freqThreshold, the bigram is skipped.
	"""
	corpusLength, uniModel, biModel = _getModels(corpusName)
	freqAndBi = zip(biModel.values(), biModel.keys())
	freqAndBi.sort()
	freqAndBi.reverse()
	if freqThreshold:
		freqThreshold = int(corpusLength/float(freqThreshold))
	for _, bigram in freqAndBi:
		head, tail = bigram
		if freqThreshold and (uniModel[head]>freqThreshold
			or uniModel[tail]>freqThreshold):
			continue
		chisq, valid = computeChisquare([
			[biModel[head,tail], uniModel[head]-biModel[head,tail]],
			[uniModel[tail]-biModel[head,tail], len(biModel)-
				uniModel[head]-uniModel[tail]+biModel[head,tail]]
			])
		pVal = lchisqprob(chisq, 1)
		if pVal<niveau:
			prefix = "ja:"
		else:
			prefix = "    nein:"
		if valid:
			print prefix+"%s %s -- %f"%(bigram+(pVal,))
		else:
			print prefix+"(%s %s)"%bigram


if __name__=="__main__":
	import sys, locale
	locale.setlocale(locale.LC_ALL, "")
	#runCollocTest(sys.argv[1])
	#showSortedCollocs(sys.argv[1], niveau=0.01, freqThreshold=0)
